import { Observable } from 'rxjs/Observable';
import { Observer } from 'rxjs/Observer';
import { Subject } from 'rxjs/Subject';
import {
  WebSocketSubject,
  WebSocketSubjectConfig
} from 'rxjs/observable/dom/WebSocketSubject';

import 'rxjs/add/operator/distinctUntilChanged';
import 'rxjs/add/operator/share';
import 'rxjs/add/operator/takeWhile';
import 'rxjs/add/observable/interval';

export interface Codec {
  decode(e: MessageEvent): any;
  encode(data: any): string;
}

export const defaultCodec: Codec = {
  decode: (e: MessageEvent) => {
    return JSON.parse(e.data);
  },
  encode: (data: any): string => {
    return JSON.stringify(data);
  }
};

export class RxWebsocketSubject<T> extends Subject<T> {
  private reconnectionObservable: Observable<number>;
  private wsSubjectConfig: WebSocketSubjectConfig;
  private socket: WebSocketSubject<any>;
  private connectionObserver: Observer<boolean>;
  private _connectionStatus: Observable<boolean>;

  private _reconnectInterval = 5000;
  private _reconnectAttempts = 10;

  public constructor(private url: string, private codec?: Codec) {
    super();

    this._connectionStatus = new Observable<boolean>((observer) => {
      this.connectionObserver = observer;
    }).share().distinctUntilChanged();

    if (undefined === codec) {
      this.codec = defaultCodec;
    }

    this.wsSubjectConfig = {
      url: url,
      closeObserver: {
        next: (e: CloseEvent) => {
          this.socket = null;
          this.connectionObserver.next(false);
        }
      },
      openObserver: {
        next: (e: Event) => {
          this.connectionObserver.next(true);
        }
      },
      // resultSelector: this.codec.decode,
    };

    this._connectionStatus.subscribe((isConnected: boolean) => {
      if (!this.reconnectionObservable && typeof(isConnected) === 'boolean' && !isConnected) {
        this.reconnect();
      }
    });
  }

  public set reconnectInterval(reconnectInterval: number) {
    this._reconnectInterval = reconnectInterval;
  }

  public set reconnectAttempts(reconnectAttempts: number) {
    this._reconnectAttempts = reconnectAttempts;
  }

  public get connectionStatus(): Observable<boolean> {
    return this._connectionStatus;
  }

  public connect(): void {
    this.socket = new WebSocketSubject(this.wsSubjectConfig);
    this.socket.subscribe(
      (m) => {
        this.next(m);
      },
      (error: Event) => {
        if (!this.socket) {
          this.reconnect();
        }
      }
    );
  }

  private reconnect(): void {
    this.reconnectionObservable = Observable.interval(this._reconnectInterval)
      .takeWhile((v, index) => {
        return index < this._reconnectAttempts && !this.socket;
    });
    this.reconnectionObservable.subscribe(
      () => {
        this.connect();
      },
      null,
      () => {
        this.reconnectionObservable = null;
        if (!this.socket) {
          this.complete();
          this.connectionObserver.complete();
        }
      }
    );
  }

  public send(data: any): void {
    const s = this.codec.encode(data);
    this.socket.next(this.codec.encode(data));
  }
}
