import { Observable } from 'rxjs/Observable';
import { Subject } from 'rxjs/Subject';

import { RPCClientRWC } from './rwc/RPCClientRWC';
import { RPCRegistry } from './../registry/RPCRegistry';
import { RPCClientCodec, RPCClientResponseCodec } from '../protocol/RPCClientCodec';

export class RPCClient {
  private _requestID: number;

  private _pendingRequestsCount: number;
  private _pendingRequests: Map<number, Subject<any>>;

  public constructor(
    private _codec: RPCClientCodec,
    private _rwc: RPCClientRWC,
  ) {
    this._requestID = 0;
    this._pendingRequestsCount = 0;
    this._pendingRequests = new Map();
  }

  private getRequestID(): number {
    return ++this._requestID;
  }

  /**
   * connect
   */
  public connect(queryString?: string): void {
    this._rwc.connect(queryString);
    this._rwc.readResponse().subscribe(
      (value: Object) => {
        this.onMessage(value);
      },
      (error: any) => {
      },
      () => {

      }
    );
  }

  /**
   * close
   */
  public disconnect() {
    this._rwc.disconnect();
  }

  /**
   * notify
   */
  public notify(method: string, ...args: any[]): void {
    this.send(false, method, args);
  }

  /**
   * call
   */
  public call<T>(method: string, ...args: any[]): Observable<T> {
    return this.send<T>(true, method, args);
  }

  /**
   * callTimeout
   */
  public callTimeout<T>(ms: number, method: string, ...args: any[]): Observable<T> {

    return undefined;
  }

  private send<T>(hasResponse: boolean, method: string, args?: any[]): Observable<T> | undefined {
    let id: number;
    let resSubject: Subject<T>;
    if (hasResponse) {
      id = this.getRequestID();
      resSubject = new Subject<T>();
      this._pendingRequests.set(id, resSubject);
      this._pendingRequestsCount++;
    }

    const req = this._codec.request(method, args, id);
    this._rwc.writeRequest(req);

    if (undefined !== resSubject) {
      return resSubject.asObservable();
    }
  }

  private onMessage(message: Object): void {
    const resCodec = this._codec.response(message);
    if (undefined !== resCodec.id()) {
      this.onResponse(resCodec);
    }
  }

  protected onResponse(resCodec: RPCClientResponseCodec): void {
    const id = resCodec.id();
    const result = resCodec.result();
    const error = resCodec.error();

    const resSubject: Subject<any> = this._pendingRequests.get(id);
    this._pendingRequests.delete(id);
    this._pendingRequestsCount--;

    if (undefined !== result) {
      resSubject.next(result);
    } else if (undefined !== error) {
      resSubject.error(error);
    }
  }
}
